
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import log

from keystone.common import sql
import keystone.conf
from keystone.server import backends


CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)


def configure(version=None, config_files=None,
              pre_setup_logging_fn=lambda: None):
    keystone.conf.configure()
    sql.initialize()
    keystone.conf.set_config_defaults()

    CONF(project='keystone', version=version,
         default_config_files=config_files)

    pre_setup_logging_fn()
    keystone.conf.setup_logging()

    if CONF.insecure_debug:
        LOG.warning(
            'insecure_debug is enabled so responses may include sensitive '
            'information.')


def setup_backends(load_extra_backends_fn=lambda: {},
                   startup_application_fn=lambda: None):
    drivers = backends.load_backends()
    drivers.update(load_extra_backends_fn())
    res = startup_application_fn()
    return drivers, res



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import os
import stat

from cryptography import fernet
from oslo_log import log

from keystone.common import utils
import keystone.conf


LOG = log.getLogger(__name__)

CONF = keystone.conf.CONF

# NOTE(lbragstad): In the event there are no encryption keys on disk, let's use
# a default one until a proper key repository is set up. This allows operators
# to gracefully upgrade from Mitaka to Newton without a key repository,
# especially in multi-node deployments. The NULL_KEY is specific to credential
# encryption only and has absolutely no beneficial purpose outside of easing
# upgrades.
NULL_KEY = base64.urlsafe_b64encode(b'\x00' * 32)


class FernetUtils(object):

    def __init__(self, key_repository, max_active_keys,
                 config_group):
        self.key_repository = key_repository
        self.max_active_keys = max_active_keys
        self.config_group = config_group

    def validate_key_repository(self, requires_write=False):
        """Validate permissions on the key repository directory."""
        # NOTE(lbragstad): We shouldn't need to check if the directory was
        # passed in as None because we don't set allow_no_values to True.

        # ensure current user has sufficient access to the key repository
        is_valid = (os.access(self.key_repository, os.R_OK) and
                    os.access(self.key_repository, os.X_OK))
        if requires_write:
            is_valid = (is_valid and
                        os.access(self.key_repository, os.W_OK))

        if not is_valid:
            LOG.error(
                'Either [%(config_group)s] key_repository does not exist '
                'or Keystone does not have sufficient permission to '
                'access it: %(key_repo)s',
                {'key_repo': self.key_repository,
                 'config_group': self.config_group})
        else:
            # ensure the key repository isn't world-readable
            stat_info = os.stat(self.key_repository)
            if(stat_info.st_mode & stat.S_IROTH or
               stat_info.st_mode & stat.S_IXOTH):
                LOG.warning(
                    'key_repository is world readable: %s',
                    self.key_repository)

        return is_valid

    def create_key_directory(self, keystone_user_id=None,
                             keystone_group_id=None):
        """Attempt to create the key directory if it doesn't exist."""
        utils.create_directory(
            self.key_repository, keystone_user_id=keystone_user_id,
            keystone_group_id=keystone_group_id
        )

    def _create_new_key(self, keystone_user_id, keystone_group_id):
        """Securely create a new encryption key.

        Create a new key that is readable by the Keystone group and Keystone
        user.

        To avoid disk write failure, this function will create a tmp key file
        first, and then rename it as the valid new key.
        """
        self._create_tmp_new_key(keystone_user_id, keystone_group_id)
        self._become_valid_new_key()

    def _create_tmp_new_key(self, keystone_user_id, keystone_group_id):
        """Securely create a new tmp encryption key.

        This created key is not effective until _become_valid_new_key().
        """
        key = fernet.Fernet.generate_key()  # key is bytes

        # This ensures the key created is not world-readable
        old_umask = os.umask(0o177)
        if keystone_user_id and keystone_group_id:
            old_egid = os.getegid()
            old_euid = os.geteuid()
            os.setegid(keystone_group_id)
            os.seteuid(keystone_user_id)
        elif keystone_user_id or keystone_group_id:
            LOG.warning(
                'Unable to change the ownership of the new key without a '
                'keystone user ID and keystone group ID both being provided: '
                '%s', self.key_repository)
        # Determine the file name of the new key
        key_file = os.path.join(self.key_repository, '0.tmp')
        create_success = False
        try:
            with open(key_file, 'w') as f:
                # convert key to str for the file.
                f.write(key.decode('utf-8'))
                f.flush()
                create_success = True
        except IOError:
            LOG.error('Failed to create new temporary key: %s', key_file)
            raise
        finally:
            # After writing the key, set the umask back to it's original value.
            # Do the same with group and user identifiers if a Keystone group
            # or user was supplied.
            os.umask(old_umask)
            if keystone_user_id and keystone_group_id:
                os.seteuid(old_euid)
                os.setegid(old_egid)
            # Deal with the tmp key file
            if not create_success and os.access(key_file, os.F_OK):
                os.remove(key_file)

        LOG.info('Created a new temporary key: %s', key_file)

    def _become_valid_new_key(self):
        """Make the tmp new key a valid new key.

        The tmp new key must be created by _create_tmp_new_key().
        """
        tmp_key_file = os.path.join(self.key_repository, '0.tmp')
        valid_key_file = os.path.join(self.key_repository, '0')

        os.rename(tmp_key_file, valid_key_file)

        LOG.info('Become a valid new key: %s', valid_key_file)

    def _get_key_files(self, key_repo):
        key_files = dict()
        keys = dict()
        for filename in os.listdir(key_repo):
            path = os.path.join(key_repo, str(filename))
            if os.path.isfile(path):
                with open(path, 'r') as key_file:
                    try:
                        key_id = int(filename)
                    except ValueError:  # nosec : name is not a number
                        pass
                    else:
                        key = key_file.read()
                        if len(key) == 0:
                            LOG.warning('Ignoring empty key found in key '
                                        'repository: %s', path)
                            continue
                        key_files[key_id] = path
                        keys[key_id] = key
        return key_files, keys

    def initialize_key_repository(self, keystone_user_id=None,
                                  keystone_group_id=None):
        """Create a key repository and bootstrap it with a key.

        :param keystone_user_id: User ID of the Keystone user.
        :param keystone_group_id: Group ID of the Keystone user.

        """
        # make sure we have work to do before proceeding
        if os.access(os.path.join(self.key_repository, '0'),
                     os.F_OK):
            LOG.info('Key repository is already initialized; aborting.')
            return

        # bootstrap an existing key
        self._create_new_key(keystone_user_id, keystone_group_id)

        # ensure that we end up with a primary and secondary key
        self.rotate_keys(keystone_user_id, keystone_group_id)

    def rotate_keys(self, keystone_user_id=None, keystone_group_id=None):
        """Create a new primary key and revoke excess active keys.

        :param keystone_user_id: User ID of the Keystone user.
        :param keystone_group_id: Group ID of the Keystone user.

        Key rotation utilizes the following behaviors:

        - The highest key number is used as the primary key (used for
          encryption).
        - All keys can be used for decryption.
        - New keys are always created as key "0," which serves as a placeholder
          before promoting it to be the primary key.

        This strategy allows you to safely perform rotation on one node in a
        cluster, before syncing the results of the rotation to all other nodes
        (during both key rotation and synchronization, all nodes must recognize
        all primary keys).

        """
        # read the list of key files
        key_files, _ = self._get_key_files(self.key_repository)

        LOG.info('Starting key rotation with %(count)s key files: '
                 '%(list)s', {
                     'count': len(key_files),
                     'list': list(key_files.values())})

        # add a tmp new key to the rotation, which will be the *next* primary
        self._create_tmp_new_key(keystone_user_id, keystone_group_id)

        # determine the number of the new primary key
        current_primary_key = max(key_files.keys())
        LOG.info('Current primary key is: %s', current_primary_key)
        new_primary_key = current_primary_key + 1
        LOG.info('Next primary key will be: %s', new_primary_key)

        # promote the next primary key to be the primary
        os.rename(
            os.path.join(self.key_repository, '0'),
            os.path.join(self.key_repository, str(new_primary_key))
        )
        key_files.pop(0)
        key_files[new_primary_key] = os.path.join(
            self.key_repository,
            str(new_primary_key))
        LOG.info('Promoted key 0 to be the primary: %s', new_primary_key)

        # rename the tmp key to the real staged key
        self._become_valid_new_key()

        max_active_keys = self.max_active_keys

        # purge excess keys

        # Note that key_files doesn't contain the new active key that was
        # created, only the old active keys.
        keys = sorted(key_files.keys(), reverse=True)
        while len(keys) > (max_active_keys - 1):
            index_to_purge = keys.pop()
            key_to_purge = key_files[index_to_purge]
            LOG.info('Excess key to purge: %s', key_to_purge)
            os.remove(key_to_purge)

    def load_keys(self, use_null_key=False):
        """Load keys from disk into a list.

        The first key in the list is the primary key used for encryption. All
        other keys are active secondary keys that can be used for decrypting
        tokens.

        :param use_null_key: If true, a known key containing null bytes will be
                             appended to the list of returned keys.

        """
        if not self.validate_key_repository():
            if use_null_key:
                return [NULL_KEY]
            return []

        # build a dictionary of key_number:encryption_key pairs
        _, keys = self._get_key_files(self.key_repository)

        if len(keys) != self.max_active_keys:
            # Once the number of keys matches max_active_keys, this log entry
            # is too repetitive to be useful. Also note that it only makes
            # sense to log this message for tokens since credentials doesn't
            # have a `max_active_key` configuration option.
            if self.key_repository == CONF.fernet_tokens.key_repository:
                msg = ('Loaded %(count)d Fernet keys from %(dir)s, but '
                       '`[fernet_tokens] max_active_keys = %(max)d`; perhaps '
                       'there have not been enough key rotations to reach '
                       '`max_active_keys` yet?')
                LOG.debug(msg, {
                    'count': len(keys),
                    'max': self.max_active_keys,
                    'dir': self.key_repository})

        # return the encryption_keys, sorted by key number, descending
        key_list = [keys[x] for x in sorted(keys.keys(), reverse=True)]

        if use_null_key:
            key_list.append(NULL_KEY)

        return key_list



# Copyright 2013 OpenStack Foundation
# Copyright 2013 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ldap.modlist

from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone import identity
from keystone.tests import unit
from keystone.tests.unit import test_ldap_livetest


CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


def create_object(dn, attrs):
    conn = ldap.initialize(CONF.ldap.url)
    conn.simple_bind_s(CONF.ldap.user, CONF.ldap.password)
    ldif = ldap.modlist.addModlist(attrs)
    conn.add_s(dn, ldif)
    conn.unbind_s()


class LiveTLSLDAPIdentity(test_ldap_livetest.LiveLDAPIdentity):

    def _ldap_skip_live(self):
        self.skip_if_env_not_set('ENABLE_TLS_LDAP_LIVE_TEST')

    def config_files(self):
        config_files = super(LiveTLSLDAPIdentity, self).config_files()
        config_files.append(unit.dirs.tests_conf('backend_tls_liveldap.conf'))
        return config_files

    def test_tls_certfile_demand_option(self):
        self.config_fixture.config(group='ldap',
                                   use_tls=True,
                                   tls_cacertdir=None,
                                   tls_req_cert='demand')
        PROVIDERS.identity_api = identity.backends.ldap.Identity()

        user = unit.create_user(PROVIDERS.identity_api, 'default',
                                name='fake1', password='fakepass1')
        user_ref = PROVIDERS.identity_api.get_user(user['id'])
        self.assertEqual(user['id'], user_ref['id'])

        user['password'] = 'fakepass2'
        PROVIDERS.identity_api.update_user(user['id'], user)

        PROVIDERS.identity_api.delete_user(user['id'])
        self.assertRaises(
            exception.UserNotFound,
            PROVIDERS.identity_api.get_user,
            user['id']
        )

    def test_tls_certdir_demand_option(self):
        self.config_fixture.config(group='ldap',
                                   use_tls=True,
                                   tls_cacertdir=None,
                                   tls_req_cert='demand')
        PROVIDERS.identity_api = identity.backends.ldap.Identity()

        user = unit.create_user(PROVIDERS.identity_api, 'default',
                                id='fake1', name='fake1',
                                password='fakepass1')
        user_ref = PROVIDERS.identity_api.get_user('fake1')
        self.assertEqual('fake1', user_ref['id'])

        user['password'] = 'fakepass2'
        PROVIDERS.identity_api.update_user('fake1', user)

        PROVIDERS.identity_api.delete_user('fake1')
        self.assertRaises(
            exception.UserNotFound,
            PROVIDERS.identity_api.get_user,
            'fake1'
        )

    def test_tls_bad_certfile(self):
        self.config_fixture.config(
            group='ldap',
            use_tls=True,
            tls_req_cert='demand',
            tls_cacertfile='/etc/keystone/ssl/certs/mythicalcert.pem',
            tls_cacertdir=None)
        PROVIDERS.identity_api = identity.backends.ldap.Identity()

        user = unit.new_user_ref('default')
        self.assertRaises(IOError, PROVIDERS.identity_api.create_user, user)

    def test_tls_bad_certdir(self):
        self.config_fixture.config(
            group='ldap',
            use_tls=True,
            tls_cacertfile=None,
            tls_req_cert='demand',
            tls_cacertdir='/etc/keystone/ssl/mythicalcertdir')
        PROVIDERS.identity_api = identity.backends.ldap.Identity()

        user = unit.new_user_ref('default')
        self.assertRaises(IOError, PROVIDERS.identity_api.create_user, user)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import textwrap

import pycodestyle

from keystone.tests.hacking import checks
from keystone.tests import unit
from keystone.tests.unit.ksfixtures import hacking as hacking_fixtures


class BaseStyleCheck(unit.BaseTestCase):

    def setUp(self):
        super(BaseStyleCheck, self).setUp()
        self.code_ex = self.useFixture(self.get_fixture())
        self.addCleanup(delattr, self, 'code_ex')

    def get_checker(self):
        """Return the checker to be used for tests in this class."""
        raise NotImplementedError('subclasses must provide '
                                  'a real implementation')

    def get_fixture(self):
        return hacking_fixtures.HackingCode()

    def run_check(self, code):
        pycodestyle.register_check(self.get_checker())

        lines = textwrap.dedent(code).strip().splitlines(True)

        # Load all keystone hacking checks, they are of the form Kddd,
        # where ddd can from range from 000-999
        guide = pycodestyle.StyleGuide(select='K')
        checker = pycodestyle.Checker(lines=lines, options=guide.options)
        checker.check_all()
        checker.report._deferred_print.sort()
        return checker.report._deferred_print

    def assert_has_errors(self, code, expected_errors=None):
        actual_errors = [e[:3] for e in self.run_check(code)]
        self.assertCountEqual(expected_errors or [], actual_errors)


class TestCheckForMutableDefaultArgs(BaseStyleCheck):

    def get_checker(self):
        return checks.CheckForMutableDefaultArgs

    def test(self):
        code = self.code_ex.mutable_default_args['code']
        errors = self.code_ex.mutable_default_args['expected_errors']
        self.assert_has_errors(code, expected_errors=errors)


class TestBlockCommentsBeginWithASpace(BaseStyleCheck):

    def get_checker(self):
        return checks.block_comments_begin_with_a_space

    def test(self):
        code = self.code_ex.comments_begin_with_space['code']
        errors = self.code_ex.comments_begin_with_space['expected_errors']
        self.assert_has_errors(code, expected_errors=errors)


class TestTranslationChecks(BaseStyleCheck):

    def get_checker(self):
        return checks.CheckForTranslationIssues

    def get_fixture(self):
        return hacking_fixtures.HackingTranslations()

    def assert_has_errors(self, code, expected_errors=None):
        # pull out the parts of the error that we'll match against
        actual_errors = (e[:3] for e in self.run_check(code))
        # adjust line numbers to make the fixture data more readable.
        import_lines = len(self.code_ex.shared_imports.split('\n')) - 1
        actual_errors = [(e[0] - import_lines, e[1], e[2])
                         for e in actual_errors]
        self.assertEqual(expected_errors or [], actual_errors)

    def test_for_translations(self):
        for example in self.code_ex.examples:
            code = self.code_ex.shared_imports + example['code']
            errors = example['expected_errors']
            self.assert_has_errors(code, expected_errors=errors)


class TestDictConstructorWithSequenceCopy(BaseStyleCheck):

    def get_checker(self):
        return checks.dict_constructor_with_sequence_copy

    def test(self):
        code = self.code_ex.dict_constructor['code']
        errors = self.code_ex.dict_constructor['expected_errors']
        self.assert_has_errors(code, expected_errors=errors)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

from keystone.common.policies import base

access_token_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'authorize_request_token',
        check_str=base.RULE_ADMIN_REQUIRED,
        # Since access tokens require a request token and request tokens
        # require a project, it makes sense to have a project-scoped token in
        # order to access these APIs.
        scope_types=['project'],
        description='Authorize OAUTH1 request token.',
        operations=[{'path': '/v3/OS-OAUTH1/authorize/{request_token_id}',
                     'method': 'PUT'}]),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_access_token',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['project'],
        description='Get OAUTH1 access token for user by access token ID.',
        operations=[{'path': ('/v3/users/{user_id}/OS-OAUTH1/access_tokens/'
                              '{access_token_id}'),
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_access_token_role',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['project'],
        description='Get role for user OAUTH1 access token.',
        operations=[{'path': ('/v3/users/{user_id}/OS-OAUTH1/access_tokens/'
                              '{access_token_id}/roles/{role_id}'),
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_access_tokens',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['project'],
        description='List OAUTH1 access tokens for user.',
        operations=[{'path': '/v3/users/{user_id}/OS-OAUTH1/access_tokens',
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_access_token_roles',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['project'],
        description='List OAUTH1 access token roles.',
        operations=[{'path': ('/v3/users/{user_id}/OS-OAUTH1/access_tokens/'
                              '{access_token_id}/roles'),
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_access_token',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['project'],
        description='Delete OAUTH1 access token.',
        operations=[{'path': ('/v3/users/{user_id}/OS-OAUTH1/access_tokens/'
                              '{access_token_id}'),
                     'method': 'DELETE'}])
]


def list_rules():
    return access_token_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# A null initial migration to open this repo. Do not re-use replace this with
# a real migration, add additional ones in subsequent version scripts.

import sqlalchemy as sql
import sqlalchemy.orm

NULL_DOMAIN_ID = '<<keystone.domain.root>>'


def upgrade(migrate_engine):

    def _generate_root_domain_project():
        # Generate a project that will act as a root for all domains, in order
        # for use to be able to use a FK constraint on domain_id. Projects
        # acting as a domain will not reference this as their parent_id, just
        # as domain_id.
        #
        # This special project is filtered out by the driver, so is never
        # visible to the manager or API.

        project_ref = {
            'id': NULL_DOMAIN_ID,
            'name': NULL_DOMAIN_ID,
            'enabled': False,
            'description': '',
            'domain_id': NULL_DOMAIN_ID,
            'is_domain': True,
            'parent_id': None,
            'extra': '{}',
        }
        return project_ref

    meta = sql.MetaData()
    meta.bind = migrate_engine
    session = sql.orm.sessionmaker(bind=migrate_engine)()

    project = sql.Table('project', meta, autoload=True)

    root_domain_project = _generate_root_domain_project()
    new_entry = project.insert().values(**root_domain_project)
    session.execute(new_entry)
    session.commit()

    session.close()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# NOTE(morgan): This entire module is to provide compatibility for the old
# @protected style decorator enforcement. All new enforcement should directly
# reference the Enforcer object itself.
from keystone.common.rbac_enforcer import enforcer
from keystone import conf


CONF = conf.CONF


# NOTE(morgan): Shared-state enforcer object
_ENFORCER = enforcer.RBACEnforcer()


def reset():
    _ENFORCER._reset()


def get_enforcer():
    """Entrypoint that must return the raw oslo.policy enforcer obj.

    This is utilized by the command-line policy tools.

    :returns: :class:`oslo_policy.policy.Enforcer`
    """
    CONF(project='keystone')
    return _ENFORCER._enforcer


enforce = _ENFORCER._enforce



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

DEPRECATED_REASON = (
    "The federated mapping API is now aware of system scope and default roles."
)

deprecated_get_mapping = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_mapping',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN
)
deprecated_list_mappings = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_mappings',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN
)
deprecated_update_mapping = policy.DeprecatedRule(
    name=base.IDENTITY % 'update_mapping',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN
)
deprecated_create_mapping = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_mapping',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN
)
deprecated_delete_mapping = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_mapping',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN
)


mapping_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_mapping',
        check_str=base.SYSTEM_ADMIN,
        # FIXME(lbragstad): Today, keystone doesn't support federation unless
        # the person create identity providers, service providers, or mappings
        # has the ability to modify keystone and Apache configuration files.
        # If, or when, keystone adds support for federating identities without
        # having to touch system configuration files, the list of `scope_types`
        # for these policies should include `project`.
        scope_types=['system'],
        description=('Create a new federated mapping containing one or '
                     'more sets of rules.'),
        operations=[{'path': '/v3/OS-FEDERATION/mappings/{mapping_id}',
                     'method': 'PUT'}],
        deprecated_rule=deprecated_create_mapping),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_mapping',
        check_str=base.SYSTEM_READER,
        scope_types=['system'],
        description='Get a federated mapping.',
        operations=[
            {
                'path': '/v3/OS-FEDERATION/mappings/{mapping_id}',
                'method': 'GET'
            },
            {
                'path': '/v3/OS-FEDERATION/mappings/{mapping_id}',
                'method': 'HEAD'
            }
        ],
        deprecated_rule=deprecated_get_mapping
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_mappings',
        check_str=base.SYSTEM_READER,
        scope_types=['system'],
        description='List federated mappings.',
        operations=[
            {
                'path': '/v3/OS-FEDERATION/mappings',
                'method': 'GET'
            },
            {
                'path': '/v3/OS-FEDERATION/mappings',
                'method': 'HEAD'
            }
        ],
        deprecated_rule=deprecated_list_mappings,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_mapping',
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Delete a federated mapping.',
        operations=[{'path': '/v3/OS-FEDERATION/mappings/{mapping_id}',
                     'method': 'DELETE'}],
        deprecated_rule=deprecated_delete_mapping),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'update_mapping',
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Update a federated mapping.',
        operations=[{'path': '/v3/OS-FEDERATION/mappings/{mapping_id}',
                     'method': 'PATCH'}],
        deprecated_rule=deprecated_update_mapping)
]


def list_rules():
    return mapping_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import uuid

from keystone.common import provider_api
from keystone.common import sql
import keystone.conf
from keystone import exception
from keystone.identity.backends import sql_model as model
from keystone.tests import unit


CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class ShadowUsersBackendTests(object):
    def test_create_nonlocal_user_unique_constraint(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_created = PROVIDERS.shadow_users_api.create_nonlocal_user(user)
        self.assertNotIn('password', user_created)
        self.assertEqual(user_created['id'], user['id'])
        self.assertEqual(user_created['domain_id'], user['domain_id'])
        self.assertEqual(user_created['name'], user['name'])
        new_user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        new_user['name'] = user['name']
        self.assertRaises(exception.Conflict,
                          PROVIDERS.shadow_users_api.create_nonlocal_user,
                          new_user)

    def test_create_nonlocal_user_does_not_create_local_user(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        new_nonlocal_user = PROVIDERS.shadow_users_api.create_nonlocal_user(
            user
        )
        user_ref = self._get_user_ref(new_nonlocal_user['id'])
        self.assertIsNone(user_ref.local_user)

    def test_nonlocal_user_unique_user_id_constraint(self):
        user_ref = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.shadow_users_api.create_nonlocal_user(user_ref)
        # attempt to create a nonlocal_user with the same user_id
        nonlocal_user = {
            'domain_id': CONF.identity.default_domain_id,
            'name': uuid.uuid4().hex,
            'user_id': user['id']
        }
        self.assertRaises(sql.DBDuplicateEntry, self._add_nonlocal_user,
                          nonlocal_user)

    def test_get_user(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user.pop('email')
        user.pop('password')
        user_created = PROVIDERS.shadow_users_api.create_nonlocal_user(user)
        self.assertEqual(user_created['id'], user['id'])
        user_found = PROVIDERS.shadow_users_api.get_user(user_created['id'])
        self.assertCountEqual(user_created, user_found)

    def test_create_federated_user_unique_constraint(self):
        user_dict = PROVIDERS.shadow_users_api.create_federated_user(
            self.domain_id, self.federated_user)
        user_dict = PROVIDERS.shadow_users_api.get_user(user_dict["id"])
        self.assertIsNotNone(user_dict["id"])
        self.assertRaises(exception.Conflict,
                          PROVIDERS.shadow_users_api.create_federated_user,
                          self.domain_id,
                          self.federated_user)

    def test_create_federated_user_domain(self):
        user = PROVIDERS.shadow_users_api.create_federated_user(
            self.domain_id, self.federated_user)
        self.assertEqual(user['domain_id'], self.domain_id)

    def test_create_federated_user_email(self):
        user = PROVIDERS.shadow_users_api.create_federated_user(
            self.domain_id, self.federated_user, self.email)
        self.assertEqual(user['email'], self.email)

    def test_get_federated_user(self):
        user_dict_create = PROVIDERS.shadow_users_api.create_federated_user(
            self.domain_id, self.federated_user)
        user_dict_get = PROVIDERS.shadow_users_api.get_federated_user(
            self.federated_user["idp_id"],
            self.federated_user["protocol_id"],
            self.federated_user["unique_id"])
        self.assertCountEqual(user_dict_create, user_dict_get)
        self.assertEqual(user_dict_create["id"], user_dict_get["id"])

    def test_update_federated_user_display_name(self):
        user_dict_create = PROVIDERS.shadow_users_api.create_federated_user(
            self.domain_id, self.federated_user)
        new_display_name = uuid.uuid4().hex
        PROVIDERS.shadow_users_api.update_federated_user_display_name(
            self.federated_user["idp_id"],
            self.federated_user["protocol_id"],
            self.federated_user["unique_id"],
            new_display_name)
        user_ref = PROVIDERS.shadow_users_api._get_federated_user(
            self.federated_user["idp_id"],
            self.federated_user["protocol_id"],
            self.federated_user["unique_id"])
        self.assertEqual(user_ref.federated_users[0].display_name,
                         new_display_name)
        self.assertEqual(user_dict_create["id"], user_ref.id)

    def test_set_last_active_at(self):
        self.config_fixture.config(group='security_compliance',
                                   disable_user_account_days_inactive=90)
        now = datetime.datetime.utcnow().date()
        password = uuid.uuid4().hex
        user = self._create_user(password)
        with self.make_request():
            user_auth = PROVIDERS.identity_api.authenticate(
                user_id=user['id'],
                password=password)
        user_ref = self._get_user_ref(user_auth['id'])
        self.assertGreaterEqual(now, user_ref.last_active_at)

    def test_set_last_active_at_when_config_setting_is_none(self):
        self.config_fixture.config(group='security_compliance',
                                   disable_user_account_days_inactive=None)
        password = uuid.uuid4().hex
        user = self._create_user(password)
        with self.make_request():
            user_auth = PROVIDERS.identity_api.authenticate(
                user_id=user['id'],
                password=password)
        user_ref = self._get_user_ref(user_auth['id'])
        self.assertIsNone(user_ref.last_active_at)

    def _add_nonlocal_user(self, nonlocal_user):
        with sql.session_for_write() as session:
            nonlocal_user_ref = model.NonLocalUser.from_dict(nonlocal_user)
            session.add(nonlocal_user_ref)

    def _create_user(self, password):
        user = {
            'name': uuid.uuid4().hex,
            'domain_id': self.domain_id,
            'enabled': True,
            'password': password
        }
        return PROVIDERS.identity_api.create_user(user)

    def _get_user_ref(self, user_id):
        with sql.session_for_read() as session:
            return session.query(model.User).get(user_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import fixture as config_fixture

from keystone.cmd import bootstrap
from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone.tests.unit import core
from keystone.tests.unit import default_fixtures
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import database

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class TestCaseWithBootstrap(core.BaseTestCase):
    """A simpler version of TestCase that uses bootstrap.

    Re-implementation of TestCase that doesn't load a bunch of fixtures by
    hand and instead uses the bootstrap process. This makes it so that our base
    tests have the same things available to us as operators after they run
    boostrap. It also makes our tests DRY and pushes setup required for
    specific tests into the actual test class, instead of pushing it into a
    generic structure that gets loaded for every test.

    """

    def setUp(self):
        self.useFixture(database.Database())
        super(TestCaseWithBootstrap, self).setUp()
        self.config_fixture = self.useFixture(config_fixture.Config(CONF))
        CONF(args=[], project='keystone')
        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'fernet_tokens',
                CONF.fernet_tokens.max_active_keys
            )
        )

        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'fernet_receipts',
                CONF.fernet_receipts.max_active_keys
            )
        )

        self.bootstrapper = bootstrap.Bootstrapper()
        self.addCleanup(provider_api.ProviderAPIs._clear_registry_instances)
        self.addCleanup(self.clean_default_domain)
        self.bootstrapper.admin_password = 'password'
        self.bootstrapper.admin_username = 'admin'
        self.bootstrapper.project_name = 'admin'
        self.bootstrapper.admin_role_name = 'admin'
        self.bootstrapper.service_name = 'keystone'
        self.bootstrapper.public_url = 'http://localhost/identity/'
        self.bootstrapper.immutable_roles = True

        try:
            PROVIDERS.resource_api.create_domain(
                default_fixtures.ROOT_DOMAIN['id'],
                default_fixtures.ROOT_DOMAIN)
        except exception.Conflict:
            pass

        self.bootstrapper.bootstrap()

    def clean_default_domain(self):
        PROVIDERS.resource_api.update_domain(
            CONF.identity.default_domain_id, {'enabled': False}
        )
        PROVIDERS.resource_api.delete_domain(CONF.identity.default_domain_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This file handles all flask-restful resources for /v3/services

import http.client

from keystone.catalog import schema
from keystone.common import provider_api
from keystone.common import rbac_enforcer
from keystone.common import validation
from keystone.server import flask as ks_flask


ENFORCER = rbac_enforcer.RBACEnforcer
PROVIDERS = provider_api.ProviderAPIs


class ServicesResource(ks_flask.ResourceBase):
    collection_key = 'services'
    member_key = 'service'

    def _get_service(self, service_id):
        ENFORCER.enforce_call(action='identity:get_service')
        return self.wrap_member(PROVIDERS.catalog_api.get_service(service_id))

    def _list_service(self):
        filters = ['type', 'name']
        ENFORCER.enforce_call(action='identity:list_services', filters=filters)
        hints = self.build_driver_hints(filters)
        refs = PROVIDERS.catalog_api.list_services(hints=hints)
        return self.wrap_collection(refs, hints=hints)

    def get(self, service_id=None):
        if service_id is not None:
            return self._get_service(service_id)
        return self._list_service()

    def post(self):
        ENFORCER.enforce_call(action='identity:create_service')
        service = self.request_body_json.get('service')
        validation.lazy_validate(schema.service_create, service)
        service = self._assign_unique_id(self._normalize_dict(service))
        ref = PROVIDERS.catalog_api.create_service(
            service['id'], service, initiator=self.audit_initiator)
        return self.wrap_member(ref), http.client.CREATED

    def patch(self, service_id):
        ENFORCER.enforce_call(action='identity:update_service')
        service = self.request_body_json.get('service')
        validation.lazy_validate(schema.service_update, service)
        self._require_matching_id(service)
        ref = PROVIDERS.catalog_api.update_service(
            service_id, service, initiator=self.audit_initiator)
        return self.wrap_member(ref)

    def delete(self, service_id):
        ENFORCER.enforce_call(action='identity:delete_service')
        return PROVIDERS.catalog_api.delete_service(
            service_id, initiator=self.audit_initiator), http.client.NO_CONTENT


class ServiceAPI(ks_flask.APIBase):
    _name = 'services'
    _import_name = __name__
    resources = [ServicesResource]
    resource_mapping = []


APIs = (ServiceAPI,)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

import http.client

from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class _SystemUserTokenTests(object):

    def test_user_can_validate_system_scoped_token(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], self.bootstrapper.reader_role_id
        )

        system_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            system=True
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=system_auth)
            system_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = system_token
            c.get('/v3/auth/tokens', headers=self.headers)

    def test_user_can_validate_domain_scoped_token(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        user = unit.new_user_ref(domain_id=domain['id'])
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=user['id'],
            domain_id=domain['id']
        )

        domain_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            domain_id=domain['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=domain_auth)
            domain_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = domain_token
            c.get('/v3/auth/tokens', headers=self.headers)

    def test_user_can_validate_project_scoped_token(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id)
        )

        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=user['id'],
            project_id=project['id']
        )

        project_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=project_auth)
            project_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = project_token
            c.get('/v3/auth/tokens', headers=self.headers)


class _SystemMemberAndReaderTokenTests(object):

    def test_user_cannot_revoke_a_system_scoped_token(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], self.bootstrapper.reader_role_id
        )

        system_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            system=True
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=system_auth)
            system_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = system_token
            c.delete(
                '/v3/auth/tokens', headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_revoke_a_domain_scoped_token(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        user = unit.new_user_ref(domain_id=domain['id'])
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=user['id'],
            domain_id=domain['id']
        )

        domain_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            domain_id=domain['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=domain_auth)
            domain_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = domain_token
            c.delete(
                '/v3/auth/tokens', headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_revoke_a_project_scoped_token(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id)
        )

        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=user['id'],
            project_id=project['id']
        )

        project_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=project_auth)
            project_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = project_token
            c.delete(
                '/v3/auth/tokens', headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )


class SystemReaderTests(base_classes.TestCaseWithBootstrap,
                        common_auth.AuthTestMixin,
                        _SystemUserTokenTests,
                        _SystemMemberAndReaderTokenTests):

    def setUp(self):
        super(SystemReaderTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(
            system_reader
        )['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id, password=system_reader['password'],
            system=True
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemMemberTests(base_classes.TestCaseWithBootstrap,
                        common_auth.AuthTestMixin,
                        _SystemUserTokenTests,
                        _SystemMemberAndReaderTokenTests):

    def setUp(self):
        super(SystemMemberTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(
            system_reader
        )['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id, password=system_reader['password'],
            system=True
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemAdminTests(base_classes.TestCaseWithBootstrap,
                       common_auth.AuthTestMixin,
                       _SystemUserTokenTests):

    def setUp(self):
        super(SystemAdminTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_revoke_a_system_scoped_token(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], self.bootstrapper.reader_role_id
        )

        system_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            system=True
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=system_auth)
            system_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = system_token
            c.delete('/v3/auth/tokens', headers=self.headers)

    def test_user_can_revoke_a_domain_scoped_token(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        user = unit.new_user_ref(domain_id=domain['id'])
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=user['id'],
            domain_id=domain['id']
        )

        domain_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            domain_id=domain['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=domain_auth)
            domain_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = domain_token
            c.delete('/v3/auth/tokens', headers=self.headers)

    def test_user_can_revoke_a_project_scoped_token(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id)
        )

        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=user['id'],
            project_id=project['id']
        )

        project_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=project_auth)
            project_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = project_token
            c.delete('/v3/auth/tokens', headers=self.headers)


class _DomainAndProjectUserTests(object):

    def test_user_can_validate_their_own_tokens(self):
        with self.test_client() as c:
            self.headers['X-Subject-Token'] = self.token_id
            c.get('/v3/auth/tokens', headers=self.headers)

    def test_user_can_revoke_their_own_tokens(self):
        with self.test_client() as c:
            self.headers['X-Subject-Token'] = self.token_id
            c.delete('/v3/auth/tokens', headers=self.headers)

    def test_user_cannot_validate_system_scoped_token(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], self.bootstrapper.reader_role_id
        )

        system_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            system=True
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=system_auth)
            system_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = system_token
            c.get(
                '/v3/auth/tokens', headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_revoke_system_scoped_token(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], self.bootstrapper.reader_role_id
        )

        system_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            system=True
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=system_auth)
            system_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = system_token
            c.delete(
                '/v3/auth/tokens', headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_validate_domain_scoped_token(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        user = unit.new_user_ref(domain_id=domain['id'])
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=user['id'],
            domain_id=domain['id']
        )

        domain_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            domain_id=domain['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=domain_auth)
            domain_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = domain_token
            c.get(
                '/v3/auth/tokens', headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_revoke_domain_scoped_token(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        user = unit.new_user_ref(domain_id=domain['id'])
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=user['id'],
            domain_id=domain['id']
        )

        domain_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            domain_id=domain['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=domain_auth)
            domain_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = domain_token
            c.delete(
                '/v3/auth/tokens', headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_validate_project_scoped_token(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id)
        )

        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=user['id'],
            project_id=project['id']
        )

        project_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=project_auth)
            project_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = project_token
            c.get(
                '/v3/auth/tokens', headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_revoke_project_scoped_token(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id)
        )

        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user['id'] = PROVIDERS.identity_api.create_user(user)['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=user['id'],
            project_id=project['id']
        )

        project_auth = self.build_authentication_request(
            user_id=user['id'], password=user['password'],
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=project_auth)
            project_token = r.headers['X-Subject-Token']

        with self.test_client() as c:
            self.headers['X-Subject-Token'] = project_token
            c.delete(
                '/v3/auth/tokens', headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )


class DomainUserTests(base_classes.TestCaseWithBootstrap,
                      common_auth.AuthTestMixin,
                      _DomainAndProjectUserTests):

    def setUp(self):
        super(DomainUserTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_user = unit.new_user_ref(domain_id=self.domain_id)
        self.domain_user_id = PROVIDERS.identity_api.create_user(
            domain_user
        )['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=self.domain_user_id,
            domain_id=self.domain_id
        )

        auth = self.build_authentication_request(
            user_id=self.domain_user_id, password=domain_user['password'],
            domain_id=self.domain_id
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectUserTests(base_classes.TestCaseWithBootstrap,
                       common_auth.AuthTestMixin,
                       _DomainAndProjectUserTests):

    def setUp(self):
        super(ProjectUserTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']

        project_reader = unit.new_user_ref(domain_id=self.domain_id)
        project_reader_id = PROVIDERS.identity_api.create_user(
            project_reader
        )['id']
        project = unit.new_project_ref(domain_id=self.domain_id)
        project_id = PROVIDERS.resource_api.create_project(
            project['id'], project
        )['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=project_reader_id,
            project_id=project_id
        )

        auth = self.build_authentication_request(
            user_id=project_reader_id,
            password=project_reader['password'],
            project_id=project_id
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}
